#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from decimal import *
import re

def get_digits(self):
  return self.as_tuple().digits

def get_number_of_digits(self):
  return len(self.as_tuple().digits)

def get_exponent(self):
  return self.as_tuple().exponent

def format_like(a, b):
  """
  Reformat "a" to have the same number of decimal places and the same exponent
  as "b".

  Both strings are assumed to be in floating point format and a is assumed to
  have more lower decimal places
  """
  r = re.compile(r'^([+-]?)(\d+?)(?:\.(\d+?))?(?:([eE])([+-]?)(\d+))?$')
  m = r.match(a)
  n = r.match(b)
  if m and n:
    a_sign, a_before, a_after, a_e, a_e_sign, a_exp = m.groups()
    b_sign, b_before, b_after, b_e, b_e_sign, b_exp = n.groups()

    if a_after is None:
      a_after = ""

    if a_exp:
      a_exp_int = int(a_e_sign + a_exp)
    else:
      a_exp_int = 0

    if b_exp:
      b_exp_int = int(b_e_sign + b_exp)
    else:
      b_exp_int = 0

    while b_exp_int > a_exp_int:
      a_after = a_before[-1] + a_after
      a_before = a_before[:-1]
      if not a_before:
        a_before = '0'
      a_exp_int += 1

    output = a_sign + a_before
    a_i = 0

    if b_after:
      a_i = len(b_after)
      output += '.' + a_after[:a_i]

    # Round last digit if necessary.
    if a_after:
      i  = int(a_after[0])
      if i >= 5:
        j = -1
        while True:
          c = output[j]
          if c == '.':
            j -= 1
            continue
          elif c in ('+', '-'):
            output = c + '1' + output[j+1:]
            break
          else:
            i = int(c)
            if i == 9:
              output[j] = '0'
              continue
            else:
              output[j] = str(i + 1)
              break

    if b_exp:
      output += a_e + b_e_sign + b_exp

    return output

  else:
    raise ValueError("%s or %s is not a floating point format" % (a,b))

################################################################################
# Math
################################################################################

# From http://docs.python.org/py3k/library/decimal.html#recipes
# 2012-10-26

def pi():
    """Compute Pi to the current precision.

    >>> print(pi())
    3.141592653589793238462643383

    """
    getcontext().prec += 2  # extra digits for intermediate steps
    three = Decimal(3)      # substitute "three=3.0" for regular floats
    lasts, t, s, n, na, d, da = 0, three, 3, 1, 0, 0, 24
    while s != lasts:
        lasts = s
        n, na = n+na, na+8
        d, da = d+da, da+32
        t = (t * n) / d
        s += t
    getcontext().prec -= 2
    return +s               # unary plus applies the new precision

def exp(x):
    """Return e raised to the power of x.  Result type matches input type.

    >>> print(exp(Decimal(1)))
    2.718281828459045235360287471
    >>> print(exp(Decimal(2)))
    7.389056098930650227230427461
    >>> print(exp(2.0))
    7.38905609893
    >>> print(exp(2+0j))
    (7.38905609893+0j)

    """
    getcontext().prec += 2
    i, lasts, s, fact, num = 0, 0, 1, 1, 1
    while s != lasts:
        lasts = s
        i += 1
        fact *= i
        num *= x
        s += num / fact
    getcontext().prec -= 2
    return +s

def cos(x):
    """Return the cosine of x as measured in radians.

    The Taylor series approximation works best for a small value of x.
    For larger values, first compute x = x % (2 * pi).

    >>> print(cos(Decimal('0.5')))
    0.8775825618903727161162815826
    >>> print(cos(0.5))
    0.87758256189
    >>> print(cos(0.5+0j))
    (0.87758256189+0j)

    """
    getcontext().prec += 2
    i, lasts, s, fact, num, sign = 0, 0, 1, 1, 1, 1
    while s != lasts:
        lasts = s
        i += 2
        fact *= i * (i-1)
        num *= x * x
        sign *= -1
        s += num / fact * sign
    getcontext().prec -= 2
    return +s

def sin(x):
    """Return the sine of x as measured in radians.

    The Taylor series approximation works best for a small value of x.
    For larger values, first compute x = x % (2 * pi).

    >>> print(sin(Decimal('0.5')))
    0.4794255386042030002732879352
    >>> print(sin(0.5))
    0.479425538604
    >>> print(sin(0.5+0j))
    (0.479425538604+0j)

    """
    getcontext().prec += 2
    i, lasts, s, fact, num, sign = 1, 0, x, 1, x, 1
    while s != lasts:
        lasts = s
        i += 2
        fact *= i * (i-1)
        num *= x * x
        sign *= -1
        s += num / fact * sign
    getcontext().prec -= 2
    return +s
