#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from SciNum.decimal import *
from SciNum.units import Units, WithUnits







class DecimalWithUnits(WithUnits):
  """
  A Decimal with units
  """
  def __init__(self, value, units):
    value = Decimal(value)
    super(self.__class__, self).__init__(value, units)



  def copy(self):
    """Return a copy."""
    return DecimalWithUnits(self._value, self.units)







################################################################################
# MeasuredDecimal
################################################################################

def to_measured_decimal(value):
  """
  Convert the value to a MeasuredDecimal
  """

  if isinstance(value, UncertainDecimal):
    v = value.value
    v_gsp = v.adjusted()
    u = value.plus + value.minus
    u_gsp = u.adjusted()
    # The least significant figure occupies the decimal place spanned by the
    # range, e.g. 1.55 ±0.05 has a range of 0.1, so the measurement would have
    # been made with a precision <= 0.1, thus the value is accurate to the
    # tenths and the hundreths is the uncertain number, as would be the case
    # when measuring cm using a ruling with mm markings.
    if u == Decimal('1e%d' % u_gsp):
      u_gsp -= 1
    return MeasuredDecimal(v, precision=(1 + v_gsp - u_gsp))
  else:
    return MeasuredDecimal(value)



def get_average_measurement(values):
  """
  Calculate the average value given several repeated measurements of the same
  quantity.

  The significant figures are those up to and including the first uncertain
  value.

  Returns a MeasuredDecimal.
  """
  values = tuple(values)
  avg = sum(values)/len(values)
  msp = None
  digits = None
  p = None
  for v in values:
    if isinstance(v, MeasuredDecimal):
      v = v.display_value
    elif not isinstance(v, Decimal):
      v = Decimal(v)
    if msp == None:
      msp = v.adjusted()
      digits = list(get_digits(v))
      p = len(digits)
      continue
    elif v.adjusted() != msp:
      return MeasuredDecimal(avg, precision=0)
    else:
      v_digits = get_digits(v)
      for i in range(p):
        try:
          if digits[i] != v_digits[i]:
            break
        except IndexError:
          break
      else:
        continue
      digits = digits[:i]
      p = len(digits)
  return MeasuredDecimal(avg, precision=p+1)





class MeasuredDecimal():
  """
  A measured decimal number with significant figures.

  The value is represented internally by a Decimal object.

  Significant figures are tracked during arithmetic operations while keeping all
  figures internally. This avoids the introduction of rounding errors during
  operations while enabling a rounded value to be extracted when needed.
  """

  def __init__(self, value, precision=None):
    """
    Initialize a new value.

    "precision" is the number of significant figures in the value. If it is not
    given, all figures in the initial value will be considered significant.
    """
    if isinstance(value, self.__class__):
      if not precision:
        precision = value.precision
      value = value.value

    self.__setattr__('value', value)
    self.__setattr__('precision', precision)



  def __getattr__(self, attr):
    # Only called if the attribute does not exist.
    if attr == 'precision':
      value = self.value
      if value == None or value.is_infinite() or value.is_nan():
        return 0
      else:
        return get_number_of_digits(value)
    elif attr == 'value':
      return None
    else:
      raise AttributeError(attr)



  # Add some hooks to check input and automatically update the display value.
  def __setattr__(self, attr, value):
    initial_value = self.value
    initial_precision = self.precision

    if attr == 'precision':
      if value == None:
        self.__dict__[attr] = self.__getattr__(attr)
      else:
        if isinstance(value, int):
          self.__dict__[attr] = max(0, value)
        else:
          raise ValueError("%s ValueError: precision must be of type int, not %s" % (self.__class__.__name__, type(value)))

    elif attr == 'value':
      # Allow subclasses of Decimal to be set as values, which may be useful later.
      if isinstance(value, Decimal):
        # Ensure that the new value is a copy to avoid problems when updating
        # internal data in the value.
        value = value
      else:
        value = Decimal(value)
      self.__dict__[attr] = value

      if value.is_infinite() or value.is_nan():
        self.__dict__['precision'] = 0

    else:
      self.__dict__[attr] = value

    if self.value != initial_value or self.precision != initial_precision:
      self.update_display_value()




  def update_display_value(self):
    """
    Update the display format.

    Mostly for internal use.
    """
    self.display_value = self.get_display_value()




  def get_display_value(self):
    """
    Return a Decimal value with only significant figures.
    """
    if self.precision == 0:
      if self.value.is_nan() or self.value.is_infinite():
        return Decimal(self.value)
      else:
        return Decimal('NaN')
    else:
      sign, digits, exponent = self.value.as_tuple()
      nd = len(digits)
      p = self.precision
      # Defer to the decimal module for rounding.
      value = Decimal((sign, digits, p - nd)).to_integral_value()
      value = Decimal((sign, get_digits(value), nd + exponent - p))
      return value


  def get_greatest_significant_place(self):
    """
    Return the decimal power of the most signifant digit.
    """
    return self.value.adjusted()



  def get_least_significant_place(self):
    """
    Return the decimal power of the most signifant digit.
    """
    sign, digits, exponent = self.value.as_tuple()
    return len(digits) + exponent - self.precision



  def copy(self):
    """
    Return a (deep) copy.
    """
    return self.__class__(self.value, precision=self.precision)



  def __repr__(self):
    return self.__class__.__name__ + "('" + self.__str__() + "')"



  def __str__(self):
    return self.display_value.__str__()



  def __format__(self, fmt):
    return self.display_value.__format__(fmt)



  def get_values_for_comparison(self, other):
    """
    Return appropriate values for use in comparisons.
    """
    if self.precision > other.precision:
      p = self.precision
      self.precision = other.precision
      value = self.get_display_value()
      self.precision = p
      return value, other.display_value

    elif self.precision == other.precision:
      return self.display_value, other.display_value

    else:
      b, a = other.get_values_for_comparison(self)
      return a, b



  def __lt__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a < b
    else:
      return self.display_value < other



  def __le__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a <= b
    else:
      return self.display_value <= other



  def __eq__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a == b
    else:
      return self.display_value == other



  def __ne__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a != b
    else:
      return self.display_value != other



  def __gt__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a > b
    else:
      return self.display_value > other



  def __ge__(self, other):
    if isinstance(other, self.__class__):
      a, b = self.get_values_for_comparison(other)
      return a >= b
    else:
      return self.display_value >= b



  def __bool__(self):
    return self.display_value.__bool__()


  def __iadd__(self, other):
    if isinstance(other, UncertainDecimal):
      other = to_measured_decimal(other)
    if isinstance(other, self.__class__):
      lsp = max(self.get_least_significant_place(), other.get_least_significant_place())
      self.value += other.value
      self.precision = self.get_greatest_significant_place() - lsp
    else:
      # Consider the other number to be exact.
      a = self.get_greatest_significant_place()
      try:
        self.value += other
      except TypeError:
        return NotImplemented
      b = self.get_greatest_significant_place()
      self.precision += b - a
    return self



  def __add__(self, other):
    clone = self.copy()
    return clone.__iadd__(other)



  def __radd__(self, other):
    return self.__add__(other)



  def __isub__(self, other):
    return self.__iadd__(-other)



  def __sub__(self, other):
    return self.__add__(-other)



  def __rsub__(self, other):
    return self.__neg__().__radd__(other)



  def __imul__(self, other):
    if isinstance(other, UncertainDecimal):
      other = to_measured_decimal(other)
    if isinstance(other, self.__class__):
      self.precision = min(self.precision, other.precision)
      self.value *= other.value
    else:
      # Treat other numbers as exact.
      # 2 * x should be equal to x + x.
      a = self.get_greatest_significant_place()
      try:
        self.value *= other
      except TypeError as e:
        try:
          self.value *= Decimal(other)
        except TypeError:
          return NotImplemented
      b = self.get_greatest_significant_place()
      self.precision += b - a
    return self



  def __mul__(self, other):
    clone = self.copy()
    return clone.__imul__(other)



  def __rmul__(self, other):
    return self * other



  def __itruediv__(self, other):
    return self.__imul__(1 / other)



  def __truediv__(self, other):
    return self.__mul__(1 / other)



  def __rtruediv__(self, other):
    clone = self.copy()
    try:
      clone.value = other / clone.value
      return clone
    except TypeError:
      return NotImplemented



  def __ifloordiv__(self, other):
    if isinstance(other, self.__class__):
      precision = min(self.precision, other.precision)
      other = other.value
    else:
      precision = self.precision
    try:
      self.value //= other
      # Flooring the value to the nearest integer removes any significant
      # figures to the right of the decimal point, leaving only those to the
      # left up to the greatest significant place.
      self.precision = min(precision, self.get_greatest_significant_place())
      return self
    except TypeError:
      return NotImplemented



  def __floordiv__(self, other):
    clone = self.copy()
    return clone.__ifloordiv__(other)



  def __rfloordiv__(self, other):
    clone = self.copy()
    try:
      clone.value = other // clone.value
      clone.precision = min(clone.precision, clone.get_greatest_significant_place())
      return clone
    except TypeError:
      return NotImplemented



  def __imod__(self, other):
    # Defer significance arithmetic to previous methods to avoid additional
    # signficance calculuations
    # mod(x,y) = x - y * (x // y)
    return self.__isub__(self.__floordiv__(other).__imul__(other))



  def __mod__(self, other):
    clone = self.copy()
    return clone.__imod__(other)



  def __rmod__(self, other):
    try:
      # other - self * (other // self)
      return self.__rfloordiv__(other).__imul__(self).__rsub__(other)
    except TypeError:
      return NotImplemented



  def __ipow__(self, other):
    try:
      # Multiplication by self does not change the number of signficant figures.
      self.value **= other
      return self
    except TypeError:
      return NotImplemented



  def __pow__(self, other):
    clone = self.copy()
    return clone.__ipow__(other)




  def __neg__(self):
    clone = self.copy()
    clone.value = clone.value.__neg__()
    return clone



  def __pos__(self):
    clone = self.copy()
    clone.value = clone.value.__pos__()
    return clone



  def __abs__(self):
    clone = self.copy()
    clone.value = clone.value.copy_abs()
    clone.display_value = clone.display_value.copy_abs()
    return clone



  def __complex__(self):
    return self.display_value.__complex__()



  def __int__(self):
    return self.display_value.__int__()



  def __float__(self):
    return self.display_value.__float__()



  def __round__(self):
    return self.display_value.__round__()



  def to_uncertain_decimal(self):
    return to_uncertain_decimal(self)





class MeasuredDecimalWithUnits(WithUnits):
  """
  A MeasuredDecimal with units.
  """
  def __init__(self, value, units, precision=None):
    value = MeasuredDecimal(value, precision)
    super(self.__class__, self).__init__(value, units)



  def copy(self):
    """
    Return a copy.
    """
    return MeasuredDecimalWithUnits(self._value, self.units, self.precision)



  def to_uncertain_decimal(self):
    """
    Convert to UncertainDecimalWithUnits.
    """
    v = self._value.to_uncertain_decimal()
    return UncertainDecimalWithUnits(v, self.units)



################################################################################
# UncertainDecimal
################################################################################

def parse_uncertainty(x):
  """
  Parse a string representing an uncertain number.

  Given the lack of a standard string representations, this parser should be
  considered experiemental and likely to change in the future.

  The following formats are currently recognized:

    1) "value ±uncertainty", e.g. "15.0 ±0.5"
    2) "value {uncertainty}", e.g. "15.0 {0.5}
    3) "value {+uncertainty|-uncertainty}", e.g. "15.0 {+0.5|-0.3}"
    4) "value(uncertainty)", e.g. "15.0(5)"

  Spaces are optional except for 4.

  1,2 & 4 specify symmetric uncertainties. initially symmetric uncertainties
  may become asymmetric after certain operations such as division.

  3 specifies an asymmetric uncertainty.

  Format 4 may be followed by a trailing exponent, e.g. 1.5456(3)e-3, which
  is the same as 1.5456e-3 ±0.0003e-3, or 1.5456e-3 ±3e-6
  """
  if '(' in x and ')' in x:
    head, tail = x.split(')')
    x, uncertainty = head.split('(')
    x = x.strip()
    uncertainty = uncertainty.strip()
    d = x.find('.')
    # Insert a decimal place into the uncertainty if necessary.
    if d > -1:
      decimal_places = len(x) - d - 1
      u_len = len(uncertainty)
      if u_len >= decimal_places:
        uncertainty = uncertainty[:-decimal_places] + '.' + uncertainty[-decimal_places:]
        if uncertainty[0] == '.':
          uncertainty = '0' + uncertainty
      else:
        uncertainty = '0.' + '0' * (decimal_places - u_len) + uncertainty
    x += tail
    uncertainty += tail
    return (x, uncertainty, uncertainty)
  elif '±' in x:
    v, u = x.split('±')
    return v, u, u
  elif '{' in x and '}' in x:
    v, u = x.rstrip('}').split('{')
    if '|' in u:
      p, m = u.split('|-')
      return v, p, m
    else:
      return v, u, u
  else:
    return x, 0, 0





def to_uncertain_decimal(value):
  """
  Convert the value to an UncertainDecimal
  """

  if isinstance(value, MeasuredDecimal):
    # Reasoning:
    # The greatest total uncertainty (gtu) is the distance between two
    # graduations of the measuring device. There are three cases to consider:
    #
    # 1) The measurement is estimated to be in the middle of two graduations. In
    #    this case the maximum uncertainty on either side of the measurement is
    #    gtu/2, because the value must fall between the enclosing graduations.
    #
    # 2) The measurement is estimated to be on a graduation. Again the
    #    uncertainty on either side is gtu/2.
    #
    # 3) The measurement is estimate to be closer to one graduation but not on
    #    it. As in the previous cases the uncertainty on either side is at most
    #    gtu/2, but the graduation acts as a boundary and constrains the
    #    uncertainty on one side further. For example, if the graduations are
    #    given as units, and the measurement is estimated to be 1.3, it can be
    #    no less than 1.0 and so the uncertainty would be +0.5, -0.3.
    #
    # In all 3 cases the uncertainies may be less than that given above, but it
    # is not possible to deduce them from the figure alone. This approach thus
    # provides reasonable bounds for the uncertainty.
    v = value.display_value
    lsp = value.get_least_significant_place()
    digits = get_digits(v)
    last_digit = digits[-1]
    if last_digit in (0,5):
      return UncertainDecimal(v, plus=Decimal('5e%d' % lsp))
    elif last_digit > 5:
      return UncertainDecimal(v, plus=Decimal('%de%d' % (10-last_digit, lsp)), minus=Decimal('5e%d' % lsp))
    else:
      return UncertainDecimal(v, plus=Decimal('5e%d' % lsp), minus=Decimal('%de%d' % (last_digit, lsp)))
  else:
    return UncertainDecimal(value)






class UncertainDecimal():
  """
  A decimal number with arbitrary precision and uncertainty.

  The value is represented internally by a Decimal object.

  The value is considered to lie between an upper and a lower uncertainty.
  """

  def __init__(self, value, plus=None, minus=None):
    """
    Initialize a new value.

    "plus" is the upper uncertainty and "minus" is the lower uncertainty such
    that the following is always verified:

      value - abs(minus) <= value <= <= value + abs(plus)
    """
    if isinstance(value, self.__class__):
      if plus == None:
        plus = value.plus
      if minus == None:
        minus = value.minus
      value = value.value
    else:
      if isinstance(value, str):
        value, p, m = parse_uncertainty(value)
        if plus == None and p:
          plus = p
        if minus == None and m:
          minus = m
      if plus == None and minus != None:
        plus = minus
      elif minus == None and plus != None:
        minus = plus
      elif plus == None and minus == None:
        raise TypeError('%s TypeError: unable to determine uncertainty' % self.__class__.__name__)


    self.value = value
    self.plus = plus
    self.minus = minus



  def __getattr__(self, attr):
    if attr == 'upper_bound':
      return self.value + self.plus
    elif attr == 'lower_bound':
      return self.value - self.minus
    else:
      raise AttributeError(attr)



  def __setattr__(self, attr, value):
    if attr in ('value', 'plus', 'minus'):
      # Allow subclasses of Decimal to be set as values, which may be useful later.
      if isinstance(value, Decimal):
        # Ensure that the new value is a copy to avoid problems when updating
        # internal data in the value.
        value = value
      else:
        value = Decimal(value)

      if attr in ('plus', 'minus'):
        value = abs(value)

      self.__dict__[attr] = value

    elif attr in ('upper_bound', 'lower_bound'):
      raise AttributeError('"%s" is read-only' % attr)


    else:
      self.__dict__[attr] = value



  def copy(self):
    """
    Return a (deep) copy.
    """
    return self.__class__(self.value, plus=self.plus, minus=self.minus)



  def __repr__(self):
    return self.__class__.__name__ + "('" + self.__str__() + "')"



  def __str__(self, *args, **kwargs):
    return self.__format__('g', *args, **kwargs)



  def __format__(self, fmt, *args, **kwargs):
    return "%s {%s}" % self.format(fmt, *args, **kwargs)



  def format(self, fmt='g', homogenize=True, concise=False):
    """
    Convert the number and the uncertainty to an appropriate string
    representation.

    homogenize: If True, the uncertainty will be formatted to match the places
    shown in the number after it has been formatted.
    """
    v = self.value.__format__(fmt)
    if self.plus == 0 and self.minus == 0:
      return v, ''
    elif homogenize or concise:
      # Make sure that the uncertainties make sense for the requested output
      # format by checking if the upper and lower bounds make a difference.
      ub = self.upper_bound.__format__(fmt)
      lb = self.lower_bound.__format__(fmt)
      du = Decimal(ub) - Decimal(v)
      dl = Decimal(v) - Decimal(lb)
      p = format(du, fmt)
      p = format_like(p, v)
      m = format(dl, fmt)
      m = format_like(m, v)
      if v == ub and v == lb:
        return v, ''
      elif p == m:
        if concise:
          return v, '(%s)' % p.lstrip('+-0.')
        else:
          return v, '±%s' % p
      else:
        if concise:
          return v, '(+%s-%s)' % (p.lstrip('+-0.'), m.lstrip('+-0.'))
        else:
          return v, '+%s|-%s' % (p, m)
    else:
      p = self.plus
      pf = p.__format__(fmt)
      m = self.minus
      if p == m:
        return v, ''
      elif p == m:
        return v, '±%s' % pf
      else:
        return v, '+%s|-%s' % (pf, m.__format__(fmt))



  def __lt__(self, other):
    if isinstance(other, self.__class__):
      other = other.lower_bound
    return self.upper_bound < other



  def __le__(self, other):
    if isinstance(other, self.__class__):
      other = other.upper_bound
    return self.upper_bound <= other



  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.upper_bound >= other.lower_bound and self.lower_bound <= other.upper_bound
    else:
      return self.upper_bound >= other and self.lower_bound <= other



  def __ne__(self, other):
    return not self.__eq__(other)



  def __gt__(self, other):
    if isinstance(other, self.__class__):
      other = other.upper_bound
    return self.lower_bound > other



  def __ge__(self, other):
    if isinstance(other, self.__class__):
      other = other.lower_bound
    return self.lower_bound >= other



  def __bool__(self):
    return self.upper_bound.__bool__() or self.lower_bound.__bool__()



  def __iadd__(self, other):
    if isinstance(other, MeasuredDecimal):
      other = to_uncertain_decimal(other)
    if isinstance(other, self.__class__):
      self.value += other.value
      self.plus += other.plus
      self.minus += other.minus
    else:
      try:
        self.value += other
      except TypeError:
        return NotImplemented
    return self



  def __add__(self, other):
    clone = self.copy()
    return clone.__iadd__(other)



  def __radd__(self, other):
    return self.__add__(other)



  def __isub__(self, other):
    return self.__iadd__(-other)



  def __sub__(self, other):
    return self.__add__(-other)



  def __rsub__(self, other):
    return self.__neg__().__radd__(other)



  def __imul__(self, other):
    if isinstance(other, MeasuredDecimal):
      other = to_uncertain_decimal(other)
    if isinstance(other, self.__class__):
      u1 = self.upper_bound
      l1 = self.lower_bound
      u2 = other.upper_bound
      l2 = other.lower_bound
      self.value *= other.value
      possibilities = (u1 * u2, u1 * l2, l1 * u2, l1 * l2)
      if (u1 >= 0 and l1 <= 0) or (u2 >= 0 and l2 <= 0):
        possibilities += (0,)
      u = max(possibilities)
      l = min(possibilities)
      self.plus = u - self.value
      self.minus = self.value - l
    else:
      try:
        self.value *= other
        self.plus *= abs(other)
        self.minus *= abs(other)
      except TypeError as e:
        try:
          other = Decimal(other)
          self.value *= other
          self.plus *= abs(other)
          self.minus *= abs(other)
        except TypeError:
          return NotImplemented
    return self



  def __mul__(self, other):
    clone = self.copy()
    return clone.__imul__(other)



  def __rmul__(self, other):
    return self * other



  def __itruediv__(self, other):
    return self.__imul__(1 / other)



  def __truediv__(self, other):
    return self.__mul__(1 / other)



  def __rtruediv__(self, other):
    clone = self.copy()
    try:
      u = clone.upper_bound
      l = clone.lower_bound
      clone.value = other / clone.value
      # If the value could be zero, the uncertainty becomes infinite in both
      # directions.
      if l <= 0 and u >= 0:
        getcontext._raise_error(DivisionByZero, "divisor in range [%s,%s]" % (l, u))
        clone.plus = 'Infinity'
        clone.minus = '-Infinity'
      else:
        possibilities = (other / u, other / l)
        clone.plus = max(possibilities) - clone.value
        clone.minus = clone.value - min(possibilities)
      return clone
    except TypeError:
      return NotImplemented



  def __ifloordiv__(self, other):
    self.__itruediv__(other)
    with localcontext() as ctx:
      ctx.rounding = ROUND_FLOOR
      u = self.upper_bound
      l = self.lower_bound
      self.value = self.value.to_integral_value()
      self.plus = u.to_integral_value() - self.value
      self.minus = self.value - l.to_integral_value()



  def __floordiv__(self, other):
    clone = self.copy()
    return clone.__ifloordiv__(other)



  def __rfloordiv__(self, other):
    clone = self.copy()
    try:
      u = clone.upper_bound
      l = clone.lower_bound
      clone.value = other // clone.value
      # If the value could be zero, the uncertainty becomes infinite in both
      # directions.
      if l <= 0 and u >= 0:
        getcontext._raise_error(DivisionByZero, "divisor in range [%s,%s]" % (l, u))
        clone.plus = 'Infinity'
        clone.minus = '-Infinity'
      else:
        possibilities = (other // u, other // l)
        clone.plus = max(possibilities) - clone.value
        clone.minus = clone.value - min(possibilities)
      return clone
    except TypeError:
      return NotImplemented



  def __imod__(self, other):
    # mod(x,y) = x - y * (x // y)
    return self.__isub__(self.__floordiv__(other).__imul__(other))



  def __mod__(self, other):
    clone = self.copy()
    return clone.__imod__(other)



  def __rmod__(self, other):
    try:
      # other - self * (other // self)
      return self.__rfloordiv__(other).__imul__(self).__rsub__(other)
    except TypeError:
      return NotImplemented



  def __ipow__(self, other):
    try:
      u = self.upper_bound
      l = self.lower_bound
      self.value **= other
      possibilities = (u**other, l**other)
      if (u >= 0 and l <= 0):
        possibilities += (0,)
      self.plus = max(possibilities) - self.value
      self.minus = self.value - min(possibilities)
      return self
    except TypeError:
      return NotImplemented



  def __pow__(self, other):
    clone = self.copy()
    return clone.__ipow__(other)



  def __neg__(self):
    clone = self.copy()
    clone.plus = self.minus
    clone.minus = self.plus
    clone.value = clone.value.__neg__()
    return clone



  def __pos__(self):
    clone = self.copy()
    clone.value = clone.value.__pos__()
    clone.plus = clone.plus.__pos__()
    clone.minus = clone.minus.__pos__()
    return clone



  def __abs__(self):
    clone = self.copy()
    clone.value = clone.value.copy_abs()
    u = self.upper_bound
    l = self.lower_bound
    possibilities = (u.copy_abs(), l.copy_abs())
    if (u >= 0 and l <= 0):
      possibilities += (0,)
    clone.plus = max(possibilities) - clone.value
    clone.minus = clone.value - min(possibilities)
    return clone



  def __complex__(self):
    return self.display_value.__complex__()



  def __int__(self):
    return self.display_value.__int__()



  def __float__(self):
    return self.display_value.__float__()



  def __round__(self):
    return self.display_value.__round__()



  def to_measured_decimal(self):
    return to_measured_decimal(self)





class UncertainDecimalWithUnits(WithUnits):
  """
  An UncertainDecimal with units.
  """
  def __init__(self, value, units, plus=None, minus=None):
    value = UncertainDecimal(value, plus, minus)
    super(self.__class__, self).__init__(value, units)



  def copy(self):
    """
    Return a copy.
    """
    return UncertainDecimalWithUnits(self._value, self.units, self.plus, self.minus)



  def to_measured_decimal(self):
    """
    Convert to MeasuredDecimalWithUnits.
    """
    v = self._value.to_measured_decimal()
    return MeasuredDecimalWithUnits(v, self.units)
