#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import re
from collections import deque, namedtuple
from fractions import gcd
from SciNum.decimal import Decimal




#################################### Units #####################################

class Units(dict):
  """
  A dictionary subclass representing units and their powers.
  """

  def from_str(self, s):
    """
    Parse units from a string.
    """
    self.clear()
    if s:
      whitespace = re.compile(r'\s')
      unit = re.compile(r'^([^\^]+)(?:\^({-?\d+}|-?\d+))?$')
      for foo in whitespace.split(s):
        m = unit.match(foo)
        if m:
          u, p = m.groups()
          if not p:
            p = 1
          else:
            p = p.strip('{}')
          # Use a Decimal type just in case the powers are not integers.
          self[u] = Decimal(p)
        else:
          raise ValueError('failed to parse all units in "%s"' % s)
    return self



  def copy(self):
    """
    Return a copy.
    """
    clone = super(self.__class__, self).copy()
    return self.__class__(clone)



  def __getitem__(self, key):
    try:
      return super(self.__class__, self).__getitem__(key)
    except KeyError:
      return 0



  def __setitem__(self, key, value):
    if not isinstance(value, int):
      try:
        new_value = int(value)
        if new_value == value:
          value = new_value
      except ValueError:
        raise TypeError("unit powers must be numeric")
    if not isinstance(key, str):
      raise TypeError("unit names must be strings")
    elif value == 0:
      try:
        super(self.__class__, self).__delitem__(key)
      except KeyError:
        pass
    else:
      super(self.__class__, self).__setitem__(key, value)



  def __str__(self, braces=True):
    numerator = []
    denominator = []
    # TODO
    # Maybe use an OrderedDict to give user control over order of units.
    for unit, power in sorted(self.items()):
      if ' ' in unit:
        unit = '(' + unit + ')'
      if power != 1:
        if braces:
          unit += '^{%d}' % power
        else:
          unit += '^%d' % power
      if power > 0:
        numerator.append(unit)
      else:
        denominator.append(unit)
    return ' '.join(numerator + denominator)


  # TODO
  # Add support for `\DeclareSIUnit`
  def to_siunitx(self, wrap=True):
    """
    Return an siunitx string.

    http://ctan.org/pkg/siunitx
    """
    numerator = []
    denominator = []
    for unit, power in sorted(self.items()):

      prefix, base, fixed_prefix = get_prefix(unit)
      if fixed_prefix:
        symbol, name, indices = SI_UNITS[fixed_prefix + base]
        name = name[indices[1]:]
      else:
        symbol, name, indices = SI_UNITS[base]


      siunitx = ''
      if prefix:
        siunitx += '\\' + SI_PREFIXES[prefix].name
      siunitx += '\\' + name.replace(' ', '')
      abs_power = abs(power)
      if abs_power != 1:
        siunitx += '\\tothe{%d}' % abs(abs_power)
      if power > 0:
        numerator.append(siunitx)
      else:
        denominator.append(siunitx)
    siunitx = ''.join(numerator) + ''.join('\\per%s' % u for u in denominator)
    if wrap:
      return '\\si{%s}' % siunitx
    else:
      return siunitx



  def __repr__(self):
    numerator = []
    denominator = []
    for unit, power in sorted(self.items()):
      unit = "'%s':%d" % (unit.replace("'", "\\'"), power)
      if power > 0:
        numerator.append(unit)
      else:
        denominator.append(unit)
    return "%s({%s})" % (
      self.__class__.__name__,
      ','.join(numerator + denominator)
    )



  def __imul__(self, other):
    if isinstance(other, self.__class__):
      for k, v in other.items():
        try:
          self[k] += v
        except KeyError:
          self[k] = v
      return self
    else:
      return NotImplemented



  def __mul__(self, other):
    return self.copy().__imul__(other)



  def __itruediv__(self, other):
    if isinstance(other, self.__class__):
      for k, v in other.items():
        try:
          self[k] -= v
        except KeyError:
          self[k] = -v
      return self
    else:
      return NotImplemented



  def __truediv__(self, other):
    return self.copy().__itruediv__(other)



  def __rtruediv__(self, other):
    if isinstance(other, int) and other == 1:
      for k,v in self.items():
        self[k] = -v
      return self
    else:
      return NotImplemented



  def __ipow__(self, other):
#     if isinstance(other, int):
      for k in self:
        self[k] *= other
      return self
#     else:
#       return NotImplemented



  def __pow__(self, other):
    return self.copy().__pow__(other)





################################## namedtuple ##################################

# Prefixes along with their name and multiplication factor.
Prefix = namedtuple('Prefix', ['symbol', 'name', 'factor'])

# Units and their names. If the true base of the unit is not the same as the
# standard base (e.g. 'g' vs 'kg') then indices will be a tuple of indices that
# mark the start of the true base symbol and name, respectively.
Unit = namedtuple('Unit', ['symbol', 'name', 'indices'])

# A definition of one unit in terms of another. The factor is a proportionality
# coefficient such as the electron charge when converting J to eV. If the factor
# is 1 then it may be an integer, otherwise it should be Decimal.
Definition = namedtuple('Definition', ['symbol', 'units', 'factor'])



SI_PREFIXES = {
  'Y' : Prefix('Y', 'yotta', Decimal(10**24)),
  'Z' : Prefix('Z', 'zetta', Decimal(10**21)),
  'E' : Prefix('E', 'exa', Decimal(10**18)),
  'P' : Prefix('P', 'peta', Decimal(10**15)),
  'T' : Prefix('T', 'tera', Decimal(10**12)),
  'G' : Prefix('G', 'giga', Decimal(10**9)),
  'M' : Prefix('M', 'mega', Decimal(10**6)),
  'k' : Prefix('k', 'kilo', Decimal(10**3)),
  'h' : Prefix('h', 'hecto', Decimal(10**2)),
  'da' : Prefix('da', 'deka', Decimal(10**1)),
  'd' : Prefix('d', 'deci', Decimal(10**-1)),
  'c' : Prefix('c', 'centi', Decimal(10**-2)),
  'm' : Prefix('m', 'milli', Decimal(10**-3)),
  'µ' : Prefix('µ', 'micro', Decimal(10**-6)),
  'n' : Prefix('n', 'nano', Decimal(10**-9)),
  'p' : Prefix('p', 'pico', Decimal(10**-12)),
  'f' : Prefix('f', 'femto', Decimal(10**-15)),
  'a' : Prefix('a', 'atto', Decimal(10**-18)),
  'z' : Prefix('z', 'zepto', Decimal(10**-21)),
  'y' : Prefix('y', 'yocto', Decimal(10**-24)),
}

SI_BINARY_PREFIXES = {
  'Yi' : Prefix('Yi', 'yobi', Decimal(1024**8)),
  'Zi' : Prefix('Zi', 'zebi', Decimal(1024**7)),
  'Ei' : Prefix('Ei', 'exbi', Decimal(1024**6)),
  'Pi' : Prefix('Pi', 'pebi', Decimal(1024**5)),
  'Ti' : Prefix('Ti', 'tebi', Decimal(1024**4)),
  'Gi' : Prefix('Gi', 'gibi', Decimal(1024**3)),
  'Mi' : Prefix('Mi', 'mebi', Decimal(1024**2)),
  'ki' : Prefix('ki', 'kibi', Decimal(1024**1))
}

SI_BASE_UNITS = {
  'A' : Unit('A', 'ampere', None),
  'K' : Unit('K', 'kelvin', None),
  'cd' : Unit('cd', 'candela', None),
  'kg' : Unit('kg', 'kilogram', (1, 4)),
  'm' : Unit('m', 'meter', None),
  'mol' : Unit('mol', 'mole', None),
  's' : Unit('s', 'second', None),
}
SI_UNITS = SI_BASE_UNITS.copy()

SI_DERIVED_UNITS = {
  'Bq' : Unit('Bq', 'becquerel', None),
  '^{\circ}C' : Unit('^{\circ}C', 'degree Celsius', None),
  'C' : Unit('C', 'coulomb', None),
  'F' : Unit('F', 'farad', None),
#   'g' : Unit('g', 'gram', None),
  'Gy' : Unit('Gy', 'gray', None),
  'H' : Unit('H', 'henry', None),
  'Hz' : Unit('Hz', 'hertz', None),
  'J' : Unit('J', 'joule', None),
  'kat' : Unit('kat', 'katal', None),
  'lm' : Unit('lm', 'lumen', None),
  'lx' : Unit('lx', 'lux', None),
  'N' : Unit('N', 'newton', None),
  '\Omega' : Unit('\Omega', 'ohm', None),
  'Pa' : Unit('Pa', 'pascal', None),
  'rad' : Unit('rad', 'radian', None),
  'sr' : Unit('sr', 'steradian', None),
  'S' : Unit('S', 'simens', None),
  'Sv' : Unit('Sv', 'sievert', None),
  'T' : Unit('T', 'tesla', None),
  'V' : Unit('V', 'volt', None),
  'Wb' : Unit('Wb', 'weber', None),
  'W' : Unit('W', 'watt', None),
}

SI_UNITS.update(SI_DERIVED_UNITS)

SI_DERIVED_UNIT_DEFINITIONS = {
  'Bq' : Definition('Bq', Units({'s':-1}), 1),
  'C' : Definition('C', Units({'A':1,'s':1}), 1),
  '^{\circ}C' : Definition('^{\circ}C', Units({'K':1}), 1),
  'F' : Definition('F', Units({'C':1,'V':-1}), 1),
  # This isn't a derived unit but it faciliates insertion of kg units when
  # using reduce_to_base_units().
  'g' : Definition('g', Units({'kg':1}), Decimal('1e-3')),
  'Gy' : Definition('Gy', Units({'J':1,'kg':-1}), 1),
  'H' : Definition('H', Units({'Wb':1,'A':-1}), 1),
  'Hz' : Definition('Hz', Units({'s':-1}), 1),
  'J' : Definition('J', Units({'N':1,'m':1}), 1),
  'kat' : Definition('kat', Units({'mol':1,'s':-1}), 1),
  'lm' : Definition('lm', Units({'cd':1,'sr':1}), 1),
  'lx' : Definition('lx', Units({'lm':1,'m':-2}), 1),
  'N' : Definition('N', Units({'kg':1,'m':1,'s':-2}), 1),
  '\Omega' : Definition('\Omega', Units({'V':1,'A':-1}), 1),
  'Pa' : Definition('Pa', Units({'N':1,'m':-2}), 1),
  'rad' : Definition('rad', Units({}), 1),
  'S' : Definition('S', Units({'A':1,'V':-1}), 1),
  'sr' : Definition('sr', Units({}), 1),
  'Sv' : Definition('Sv', Units({'J':1,'kg':-1}), 1),
  'T' : Definition('T', Units({'Wb':1,'m':-2}), 1),
  'V' : Definition('V', Units({'W':1,'A':-1}), 1),
  'Wb' : Definition('Wb', Units({'V':1,'s':1}), 1),
  'W' : Definition('W', Units({'J':1,'s':-1}), 1),
}



def express(definition, unit):
  """
  Given a Definition, define a unit in terms of the other units in the
  definition.
  """
  name, units, factor = definition
  units = Units(units)
  units[name] -= 1
  power = -units[unit]
  units[unit] = 0
  if power < 0:
    for u in units:
      units[u] *= -1
  else:
    factor = 1/factor
  return Definition(unit, units, factor)


def get_prefix(
  unit,
  units=SI_UNITS,
  prefixes=SI_PREFIXES
):
  """
  Return a prefix, the base unit, and the fixed prefix of the "base" unit, e.g.
  'mg' would return ('m', 'g', 'k') because the "base" unit is 'kg'.
  """
  if unit in units:
    return None, unit, None
  for u, n, i in units.values():
    if i:
      fixed_prefix = u[:i[0]]
      u = u[i[0]:]
    if unit.endswith(u):
      prefix = unit[:-len(u)]
      if prefix in prefixes:
        return prefix, u, fixed_prefix
  else:
    return None, None, None



def remove_prefix(
  unit,
  units=SI_UNITS,
  prefixes=SI_PREFIXES
):
  """
  Remove any prefix from a unit and return the base unit and the factor.
  """
  prefix, base_unit, fixed_prefix = get_prefix(unit, units, prefixes)
  if prefix is None:
    return unit, 1
  else:
    factor = prefixes[prefix].factor
    if fixed_prefix is not None:
      factor /= prefixes[fixed_prefix].factor
    return unit, factor



def replace_names_with_symbols(
  name,
  units=SI_UNITS,
  prefixes=SI_PREFIXES
):
  """
  Replace names such as "volt" with matching symbols.

  Prefixes are also replaced, e.g. millivolt -> mV.
  """
  for u, n, i in units.values():
    if name == n:
      return n
    if i:
      true_name = n[i[1]:]
    else:
      true_name = n
    if name.endswith(true_name):
      for p, pn, f in prefixes:
        if pn + true_name == name:
          return p + u[i[0]:]
  else:
    return name



def reduce_to_base_units(
  units,
  base_units=SI_BASE_UNITS,
  definitions=SI_DERIVED_UNIT_DEFINITIONS,
  prefixes=SI_PREFIXES,
  strict=True
):
  """
  Reduce units to a set of unprefixed base units.

  Return the reduced units along with the power of 10 by which to multiply
  the value.

  The function does not currently check for cyclical definitions, so be careful.
  """
  base = Units()
  factor = 1
  definition = None
  queue = deque()
  seen = set()
  if isinstance(units, str):
    units = Units({units:1})
  elif isinstance(units, Definition):
    factor *= units.factor
    units = units.units

  for u in units.items():
    queue.append(u)

  # TODO
  # Implement cyclical definitions check.
  while queue:
    unit, power = queue.popleft()
    if unit in base_units:
      base[unit] += power
    elif unit in definitions:
      u, b, f = definitions[unit]
      for bu, bp in b.items():
        queue.append((bu, bp * power))
        factor *= f
    elif strict:
      raise ValueError('unable to reduce %s to base units with given definitions' % unit)
    else:
      base[unit] += power
  return base, factor


# TODO
# Consider removing this.

# def express_in_given_units(units, given):
#   """
#   Express in arbitrary units.
#
#   units: An iterable of tuples, each tuple containing the given unit, its power,
#   and a dictionary of its base-unit representation.
#
#   For example, to express in Hz/Pa^2, use the following:
#
#       (
#         ('Hz', 1, {'s': -1}),
#         ('Pa', -2, {'kg': 1, 'm': -1, 's': -2}),
#       )
#
#   Note that other units will appear in the output. This function simply swaps
#   equivalent (sets of) units. It does not change the dimension of the system.
#
#   The easiest way to use this function consistently is to first reduce the
#   current units to their SI base units, and then supply your own units expressed
#   in terms of the SI base units.
#   """
#   for unit, power, base in given:
#     units[unit] += power
#     for base_unit, base_power in base.items():
#       units[base_unit] -= base_power * power
#   return units



def substitute_units(units, definitions):
  """
  Substitute units with equivalent definitions.

  definitions: A list of "Definition"s to use to replace units in the list.

  This function is very useful in conjunction with reduce_to_base_units()
  for expressing quantities in an arbitrary system of units such as cgs. Simply
  reduce the units to the SI base units, then use this function to substitute
  definitions of the SI base units in the target system.
  """
  total_coeff = 1
  for unit, base, coeff in definitions:
      power = units[unit]
      if power:
        units[unit] = 0
        for u, p in base.items():
          units[u] += power * p
        total_coeff *= coeff
  return units, total_coeff




################################## WithUnits ###################################

class WithUnits():
  """A class that adds units to others."""

  def __init__(self, value, units):
    if isinstance(units, str):
      units = Units().from_str(units)
    self._value = value
    self.units = Units(units)



  def __getattr__(self, attr):
    # Pass through attributes.
    return getattr(self._value, attr)



  def __setattr__(self, attr, value):
    if attr == 'units':
      self.__dict__[attr] = Units(value)
    elif attr == '_value':
      # Decimal values are immutable
      try:
        clone = value.copy()
      except AttributeError:
        clone = value
      self.__dict__[attr] = clone
    else:
      self._value.__setattr__(attr, value)



  def copy(self):
    """
    Return a (deep) copy.
    """
    return WithUnits(self._value, self.units)



  def __repr__(self):
    return "%s(%s, %s)" % (self.__class__.__name__, self._value.__repr__(), self.units.__repr__())



  def __str__(self, *args, **kwargs):
    string = self._value.__str__(*args, **kwargs)
    if self.units:
      string += ' ' + self.units.__str__()
    return string



  def __format__(self, spec, *args, **kwargs):
    string = self._value.__format__(spec, *args, **kwargs)
    if self.units:
      string += ' ' + self.units.__str__()
    return string



  def to_siunitx(self):
    if isinstance(self._value, Decimal):
      number = '%e' % self._value
    else:
      number = self._value.to_siunitx()
    units = self.units.to_siunitx()
    return '\\Si{%s}{%s}' % (number, units)



  def reduce_to_base_units(
    self,
    base_units=SI_BASE_UNITS,
    definitions=SI_DERIVED_UNIT_DEFINITIONS,
    strict=True
  ):
    """
    Express this number in SI base units.
    """
    new_units, factor = reduce_to_base_units(self.units.copy(), base_units, definitions, strict)
    return self.__class__(self._value * factor, new_units)


#   def express_in_given_units(self, units):
#     """
#     Express this number in the given units.
#
#     See express_in_given_units().
#     """
#     new_units = express_in_given_units(self.units.copy(), units)
#     return self.__class__(self._value, new_units)


  def substitute_units(self, definitions):
    """
    Substitute the given units for others.

    See substitute_units().
    """
    new_units, coeff = substitute_units(self.units.copy(), definitions)
    return self.__class__(self._value * coeff, new_units)



  def _wc_check_unit_type(self, other):
    try:
      if self.units != other.units:
        raise TypeError("mismatched units: (%s) vs (%s)" % (self.units, other.units))
      else:
        return
    except AttributeError:
      return



  def __lt__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__lt__(other)
    else:
      return NotImplemented



  def __le__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__le__(other)
    else:
      return NotImplemented



  def __eq__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__eq__(other)
    else:
      return NotImplemented



  def __ne__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__ne__(other)
    else:
      return NotImplemented



  def __gt__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__gt__(other)
    else:
      return NotImplemented



  def __ge__(self, other):
    if isinstance(other, WithUnits):
      self._wc_check_unit_type(other)
      return self._value.__ge__(other)
    else:
      return NotImplemented



  def __iadd__(self, other):
    if isinstance(other, WithUnits):
      # The type-checking is necessary to prevent unintended conversions when
      # multiplying differnt types.
      try:
        copy = self._value.copy()
      except AttributeError:
        copy = self._value
      try:
        self._wc_check_unit_type(other)
        copy += other._value
        if isinstance(copy, self._value.__class__):
          self._value = copy
        else:
          raise TypeError
        return self
      except TypeError:
        return NotImplemented
    else:
      return NotImplemented



  def __add__(self, other):
    clone = self.copy()
    return clone.__iadd__(other)



  def __radd__(self, other):
    return self.__add__(other)



  def __isub__(self, other):
    return self.__iadd__(-other)



  def __sub__(self, other):
    return self.__add__(-other)



  def __rsub__(self, other):
    return self.__neg__().__radd__(other)



  def __imul__(self, other):
    # The type-checking is necessary to prevent unintended conversions when
    # multiplying differnt types.
    try:
      copy = self._value.copy()
    except AttributeError:
      copy = self._value
    try:
      if isinstance(other, WithUnits):
        copy *= other._value
        if isinstance(copy, self._value.__class__):
          self._value = copy
        else:
          raise TypeError
        self.units *= other.units
        return self
      else:
        copy *= other
        if isinstance(copy, self._value.__class__):
          self._value = copy
        else:
          print(type(self._value), type(other._value), type(copy))
          raise TypeError
        return self
    except TypeError:
      return NotImplemented



  def __mul__(self, other):
    clone = self.copy()
    return clone.__imul__(other)



  def __rmul__(self, other):
    return self.__mul__(other)



  def __itruediv__(self, other):
    return self.__imul__(1 / other)



  def __truediv__(self, other):
    return self.__mul__(1 / other)



  def __rtruediv__(self, other):
    try:
      clone = self.copy()
      clone.units **= (-1)
      clone._value = other / clone._value
      return clone
    except TypeError:
      return NotImplemented



  def __ifloordiv__(self, other):
    try:
      if isinstance(other, WithUnits):
        self.units /= other.units
        self._value //= other._value
      else:
        self._value //= other
      return self
    except TypeError:
      return NotImplemented



  def __floordiv__(self, other):
    clone = self.copy()
    return clone.__ifloordiv__(other)



  def __rfloordiv__(self, other):
    clone = self.copy()
    clone.units **= (-1)
    clone._value = other // clone._value
    return clone



  def __imod__(self, other):
    # mod(x,y) = x - y * (x // y)
    self.__isub__(self.__floordiv__(other).__imul__(other))
    return self



  def __mod__(self, other):
    clone = self.copy()
    return clone.__imod__(other)



  def __rmod__(self, other):
    try:
      # other - self * (other // self)
      return self.__rfloordiv__(other).__imul__(self).__rsub__(other)
    except TypeError:
      return NotImplemented



  def __ipow__(self, other):
    try:
      self.units **= other
      self._value **= other
      return self
    except TypeError:
      return NotImplemented



  def __pow__(self, other):
    clone = self.copy()
    return clone.__ipow__(other)



  def __neg__(self):
    clone = self.copy()
    clone._value = - clone._value
    return clone



  def __pos__(self):
    clone = self.copy()
    clone._value = + clone._value
    return clone



  def __abs__(self):
    clone = self.copy()
    clone._value = abs(clone._value)
    return clone



  def __complex__(self):
    return complex(self._value)



  def __int__(self):
    return int(self._value)



  def __float__(self):
    return float(self._value)



  def __round__(self):
    return round(self._value)


